#include <cppunit/TestCaller.h>
#include <cppunit/TestSuite.h>
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/extensions/TestFactoryRegistry.h>

#include "../include/vamos/geometry/Constants.h"
#include "../include/vamos/geometry/Three_Vector.h"

using namespace Vamos_Geometry;

class Null_Vector : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE (Null_Vector);
  CPPUNIT_TEST (zero_length);
  CPPUNIT_TEST_SUITE_END ();

  Three_Vector vector;

public:
  void setUp () {}
  void tearDown () {}

  void zero_length () 
  {
    CPPUNIT_ASSERT_EQUAL (vector.magnitude (), 0.0);
    CPPUNIT_ASSERT (vector.null ());
  }
};

class Unit_Vectors : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE (Unit_Vectors);
  CPPUNIT_TEST (dot_products);
  CPPUNIT_TEST (cross_products);
  CPPUNIT_TEST (add);
  CPPUNIT_TEST (subtract);
  CPPUNIT_TEST (projection);
  CPPUNIT_TEST (rotate_z);
  CPPUNIT_TEST (rotate_y);
  CPPUNIT_TEST (rotate_x);
  CPPUNIT_TEST_SUITE_END ();

  Three_Vector* mp_x;
  Three_Vector* mp_y;
  Three_Vector* mp_z;

public:
  void setUp () 
  {
    mp_x = new Three_Vector (1.0, 0.0, 0.0);
    mp_y = new Three_Vector (0.0, 1.0, 0.0);
    mp_z = new Three_Vector (0.0, 0.0, 1.0);
  }

  void tearDown () 
  {
    delete mp_x;
    delete mp_y;
    delete mp_z;
  }

  void dot_products () 
  {
    CPPUNIT_ASSERT_EQUAL (mp_x->dot (*mp_y), 0.0); 
    CPPUNIT_ASSERT_EQUAL  (mp_x->dot (*mp_z), 0.0); 
    CPPUNIT_ASSERT_EQUAL  (mp_y->dot (*mp_x), 0.0); 
    CPPUNIT_ASSERT_EQUAL  (mp_y->dot (*mp_z), 0.0); 
    CPPUNIT_ASSERT_EQUAL  (mp_z->dot (*mp_x), 0.0); 
    CPPUNIT_ASSERT_EQUAL  (mp_z->dot (*mp_y), 0.0); 
  }

  void cross_products () 
  {
    CPPUNIT_ASSERT_EQUAL  (mp_x->cross (*mp_y), *mp_z);
    CPPUNIT_ASSERT_EQUAL  (mp_x->cross (*mp_z), -*mp_y);
    CPPUNIT_ASSERT_EQUAL  (mp_y->cross (*mp_x), -*mp_z);
    CPPUNIT_ASSERT_EQUAL  (mp_y->cross (*mp_z), *mp_x);
    CPPUNIT_ASSERT_EQUAL  (mp_z->cross (*mp_x), *mp_y);
    CPPUNIT_ASSERT_EQUAL  (mp_z->cross (*mp_y), -*mp_x);
  }

  void add () 
  {
    CPPUNIT_ASSERT_EQUAL (*mp_x + *mp_y + *mp_z, Three_Vector (1.0, 1.0, 1.0));
  }

  void subtract () 
  {
    CPPUNIT_ASSERT_EQUAL (Three_Vector (1.0, 1.0, 1.0) - *mp_x - *mp_y, *mp_z);
  }

  void projection ()
  {
    CPPUNIT_ASSERT_DOUBLES_EQUAL ((mp_x->project (*mp_x)).magnitude (), 
                                  1.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL ((mp_x->project (*mp_y)).magnitude (),
                                  0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL ((mp_x->project (*mp_z)).magnitude (), 
                                  0.0, 1e-6);
  }

  void rotate_z ()
  {
    Three_Vector r = mp_x->rotate (0.0, 0.0, pi/2.0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (r.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (r.y, 1.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (r.z, 0.0, 1e-6);
  }

  void rotate_y ()
  {
    Three_Vector r = mp_z->rotate (0.0, pi/2.0, 0.0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (r.x, 1.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (r.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (r.z, 0.0, 1e-6);
  }

  void rotate_x ()
  {
    Three_Vector r = mp_y->rotate (pi/2.0, 0.0, 0.0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (r.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (r.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (r.z, 1.0, 1e-6);
  }
};

class One_Two_Three : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE (One_Two_Three);
  CPPUNIT_TEST (negate);
  CPPUNIT_TEST (zero);
  CPPUNIT_TEST (dot);
  CPPUNIT_TEST_SUITE_END ();

  Three_Vector* v123;
  Three_Vector* v321;

public:
  void setUp () 
  {
    v123 = new Three_Vector (1.0, 2.0, 3.0);
    v321 = new Three_Vector (3.0, 2.0, 1.0);
  }
  void tearDown ()
  {
    delete v123;
    delete v321;
  }

  void zero ()
  {
    v123->zero ();
    CPPUNIT_ASSERT_EQUAL (v123->magnitude (), 0.0);
  }

  void negate ()
  {
    CPPUNIT_ASSERT_EQUAL (-*v123, Three_Vector (-1.0, -2.0, -3.0));
  }

  void dot ()
  {
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v123->dot (*v321), 10.0, 1e-6);
  }
};

class X_Y_Plane : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE (X_Y_Plane);
  CPPUNIT_TEST (projection);
  CPPUNIT_TEST (back_projection);
  CPPUNIT_TEST_SUITE_END ();

  Three_Vector* x;
  Three_Vector* y;
  Three_Vector* fourty_five;

public:
  void setUp ()
  {
    x = new Three_Vector (1.0, 0.0, 0.0);
    y = new Three_Vector (0.0, 1.0, 0.0);
    fourty_five = new Three_Vector (1.0, 1.0, 0.0);
  }

  void tearDown ()
  {
    delete x;
    delete y;
    delete fourty_five;
  }

  void projection ()
  {
    Three_Vector v = x->project (*fourty_five);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v.x, 0.5, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v.y, 0.5, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v.z, 0.0, 1e-6);
  }

  void back_projection ()
  {
    Three_Vector v = x->back_project (*fourty_five);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v.x, 1.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v.y, 1.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v.z, 0.0, 1e-6);

    // Projecting and back-projecting should leave the vector
    // unchanged.
    v = x->back_project (x->project (*fourty_five));
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v.x, fourty_five->x, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v.y, fourty_five->y, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v.z, fourty_five->z, 1e-6);
  }
};

class Length_and_Angle : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE (Length_and_Angle);
  CPPUNIT_TEST (construct);
  CPPUNIT_TEST_SUITE_END ();

  Three_Vector* v;

public:
  void setUp ()
  {
    v = new Three_Vector (2.0, pi/3.0);
  }

  void tearDown ()
  {
    delete v;
  }

  void construct ()
  {
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v->x, 1.0, 1e-3);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v->y, 1.732, 1e-3);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (v->z, 0.0, 1e-3);
  }
};

CPPUNIT_TEST_SUITE_REGISTRATION (Null_Vector);
CPPUNIT_TEST_SUITE_REGISTRATION (Unit_Vectors);
CPPUNIT_TEST_SUITE_REGISTRATION (One_Two_Three);
CPPUNIT_TEST_SUITE_REGISTRATION (X_Y_Plane);
CPPUNIT_TEST_SUITE_REGISTRATION (Length_and_Angle);

int main ()
{
  CppUnit::TextUi::TestRunner runner;
  runner.addTest (CppUnit::TestFactoryRegistry::getRegistry ().makeTest ());
  runner.run ();
  return 0;
}
