#include <vamos/media/Ac3d.h>
#include <vamos/media/Texture_Image.h>
#include <vamos/geometry/Three_Vector.h>
#include <vamos/track/Gl_Road_Segment.h>

#include <GL/glut.h>

#include <cassert>

using namespace Vamos_Track;
using namespace Vamos_Geometry;
using namespace Vamos_Media;

Braking_Marker::Braking_Marker (std::string image_file,
                                double distance,
                                Side side,
                                double from_edge,
                                double off_ground,
                                double width,
                                double height)
  : mp_image (new Facade (image_file, true)),
    m_distance (distance),
    m_side (side),
    m_from_edge (from_edge),
    m_off_ground (off_ground)
{
  mp_image->set_width (width);
  mp_image->set_height (height);
}

Braking_Marker::~Braking_Marker ()
{
  delete mp_image;
}

double
Braking_Marker::width () const
{
  return mp_image->width ();
}

double
Braking_Marker::height () const
{
  return mp_image->height ();
}

void
Braking_Marker::draw () const
{
  mp_image->draw ();
}

//=============================================================================
Gl_Road_Segment::
Segment_Iterator::Segment_Iterator (double res, size_t strips)
  : texture_coordinates (Two_Vector (0.0, 0.0)),
	resolution (res)
{
}

void Gl_Road_Segment::
Segment_Iterator::start (const Three_Vector& start_coords, double start_angle)
{
  coordinates = start_coords;
  normal = Three_Vector (0.0, 0.0, 1.0);
  distance = 0.0;
  angle = start_angle;

  left_side = true;
  first = true;
  last = false;
  substrip = 0;
}

//=============================================================================
Gl_Road_Segment::Gl_Road_Segment 
(double resolution, 
 double length, 
 double radius, 
 double skew,
 const std::vector <Two_Vector>& left_width, 
 const std::vector <Two_Vector>& right_width, 
 const std::vector <Two_Vector>& left_road_width, 
 const std::vector <Two_Vector>& right_road_width, 
 Kerb* left_kerb, Kerb* right_kerb,
 double left_wall_height, double right_wall_height,
 const std::vector <Two_Vector>& elevation_points,
 const std::vector <Two_Vector>& bank_points,
 double end_bank, 
 double bank_pivot,
 const std::vector <Material>& materials,
 const std::vector <Braking_Marker*>& braking_markers)
  : Road_Segment (length, radius, 10.0, 10.0, 20.0, 20.0),
    m_gl_texture_name (0),
    m_gl_list_id (0),
    m_texture_offsets (materials.size ()),
    mp_iterator (new Segment_Iterator (resolution, materials.size ()))
{
  mp_left_width->clear ();
  mp_right_width->clear ();
  mp_left_road_width->clear ();
  mp_right_road_width->clear ();

  mp_left_width->load (left_width);
  mp_right_width->load (right_width);
  mp_left_road_width->load (left_road_width);
  mp_right_road_width->load (right_road_width);

  set_skew (skew);

  set_kerb (left_kerb, LEFT);
  set_kerb (right_kerb, RIGHT);

  set_wall_height (left_wall_height, right_wall_height);

  m_elevation_points = elevation_points;
  m_bank_points = bank_points;
  m_end_bank = end_bank;
  m_bank_pivot = bank_pivot;

  m_materials = materials;

   for (std::vector <Braking_Marker*>::const_iterator 
		 it = braking_markers.begin ();
	   it != braking_markers.end ();
	   it++)
	{
	  m_braking_markers.push_back (*it);
	}

  add_textures ();
}

Gl_Road_Segment::~Gl_Road_Segment ()
{
  delete mp_iterator;
  delete [] m_gl_texture_name;

  for (std::vector <Braking_Marker*>::iterator 
		 it = m_braking_markers.begin ();
	   it != m_braking_markers.end ();
	   it++)
	{
	  delete *it;
	}

  glDeleteLists (m_gl_list_id, 1);
  for (std::vector <GLuint>::iterator it = m_scenery_lists.begin ();
	   it != m_scenery_lists.end ();
	   it++)
	{
	  glDeleteLists (*it, 1);
	}

  for (std::vector <Texture_Image*>::iterator 
         it = m_textures.begin ();
       it != m_textures.end ();
       it++)
    {
      delete *it;
    }
}

void
Gl_Road_Segment::add_textures ()
{
  for (std::vector <Material>::iterator 
         it = m_materials.begin ();
       it != m_materials.end ();
       it++)
    {
      m_textures.push_back (new Texture_Image (it->texture_file_name (),
                                               it->smooth (),
                                               it->mip_map (),
                                               it->width (),
                                               it->height ()));
    }
}

void
Gl_Road_Segment::build ()
{
  for (std::vector <GLuint>::iterator it = m_scenery_lists.begin ();
	   it != m_scenery_lists.end ();
	   it++)
	{
	  glDeleteLists (*it, 1);
	}
  m_scenery_lists.clear ();

  for (std::vector <Model_Info>::iterator it = m_models.begin ();
	   it != m_models.end ();
	   it++)
	{
	  Ac3d model (it->file, it->scale, it->translation, 
				  deg_to_rad (1.0) * it->rotation);
	  m_scenery_lists.push_back (model.build ());
	}

  glPixelStorei (GL_UNPACK_ALIGNMENT, 1);

  if (m_gl_list_id != 0)
	{
	  glDeleteLists (m_gl_list_id, 1);
	}
  m_gl_list_id = glGenLists (1);
  glNewList (m_gl_list_id, GL_COMPILE);

  // Put the textures on a shiny white surface so we can see shading.
  glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  glEnable (GL_COLOR_MATERIAL);
  glColor3f (1.0, 1.0, 1.0);
  GLfloat specular [] = { 1.0, 1.0, 1.0, 1.0 };
  glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  GLfloat shininess [] = { 128.0 };
  glMaterialfv (GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  // Draw the left barrier.
  draw_strip (LEFT_BARRIER, m_texture_offsets [LEFT_BARRIER]);

  // Draw the left shoulder and kerb.
  glDepthMask (GL_FALSE);
  draw_strip (LEFT_SHOULDER, m_texture_offsets [LEFT_SHOULDER]);
  glDepthMask (GL_TRUE);
  draw_strip (LEFT_KERB, m_texture_offsets [LEFT_KERB]);
  glColorMask (GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
  draw_strip (LEFT_SHOULDER, m_texture_offsets [LEFT_SHOULDER]);
  glColorMask (GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

  // Draw the track.
  draw_strip (TRACK, m_texture_offsets [TRACK]);

  // Draw the right shoulder and kerb.
  glDepthMask (GL_FALSE);
  draw_strip (RIGHT_SHOULDER, m_texture_offsets [RIGHT_SHOULDER]);
  glDepthMask (GL_TRUE);
  draw_strip (RIGHT_KERB, m_texture_offsets [RIGHT_KERB]);
  glColorMask (GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
  draw_strip (RIGHT_SHOULDER, m_texture_offsets [RIGHT_SHOULDER]);
  glColorMask (GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

  // Draw the right barrier.
  draw_strip (RIGHT_BARRIER, m_texture_offsets [RIGHT_BARRIER]);
  glFlush ();

  for (std::vector <Braking_Marker*>::const_iterator 
		 it = m_braking_markers.begin ();
	   it != m_braking_markers.end ();
	   it++)
	{
	  double along = m_length - (*it)->distance ();
	  double from_center = (*it)->from_edge ();
 	  if ((*it)->side () == RIGHT)
 		{
 		  from_center = -(from_center + right_road_width (along));
 		}
 	  else
 		{
 		  from_center += left_road_width (along) + (*it)->width ();
 		}

	  double angle = m_start_angle + arc () * along / m_length;
	  double x = m_center.x - from_center * sin (angle);
	  double y = m_center.y + from_center * cos (angle);
	  if (m_radius == 0.0)
		{
		  x += along * cos (angle);
		  y += along * sin (angle);
		}
	  else
		{
 		  x += m_radius * sin (angle);
 		  y -= m_radius * cos (angle);
		}
	  double z = elevation (along, from_center) + (*it)->off_ground ();

      glPushMatrix ();
      glTranslatef (x, y, z);
      glRotatef (rad_to_deg (angle) - 90.0, 0.0, 0.0, 1.0);
      glRotatef (90.0, 1.0, 0.0, 0.0);
      (*it)->draw ();
      glPopMatrix ();
    }

  glPushMatrix ();
  glTranslatef (m_start_coords.x, m_start_coords.y, m_start_coords.z);
  glRotatef (rad_to_deg (m_start_angle), 0.0, 0.0, 1.0);
  std::for_each (m_scenery_lists.begin (), m_scenery_lists.end (), glCallList);
  glPopMatrix ();
  glEndList ();
}

void
Gl_Road_Segment::draw_strip (size_t i, double texture_offset)
{
  m_textures [i]->activate ();

  mp_iterator->start (m_start_coords, m_start_angle);
  
  Three_Vector vertex = next_vertex (i);
  glNormal3d (mp_iterator->normal.x, 
			  mp_iterator->normal.y, 
			  mp_iterator->normal.z);
  double tex_width = m_textures [i]->width (); 
  double tex_height = m_textures [i]->height ();
  double tex_x = (tex_width > 0.0)
	? mp_iterator->texture_coordinates.x / tex_width : 0.0;
  double tex_y = (tex_height > 0.0)
	? mp_iterator->texture_coordinates.y / tex_height : 0.0;
  tex_y += texture_offset;
  glTexCoord2d (tex_x, tex_y);
  glVertex3d (vertex.x, vertex.y, vertex.z);
  
  vertex = next_vertex (i);
  glNormal3d (mp_iterator->normal.x, 
			  mp_iterator->normal.y, 
			  mp_iterator->normal.z);
  tex_x = (tex_width > 0.0)
	? mp_iterator->texture_coordinates.x / tex_width : 1.0;
  glTexCoord2d (tex_x, tex_y);
  glVertex3d (vertex.x, vertex.y, vertex.z);
  
  while (!mp_iterator->last)
	{
	  vertex = next_vertex (i);
	  glNormal3d (mp_iterator->normal.x, 
				  mp_iterator->normal.y, 
				  mp_iterator->normal.z);
	  tex_x = (tex_width > 0.0)
		? mp_iterator->texture_coordinates.x / tex_width : 0.0;
	  tex_y = (tex_height > 0.0)
		? mp_iterator->texture_coordinates.y / tex_height : 1.0;
	  tex_y += texture_offset;
	  glTexCoord2d (tex_x, tex_y);
	  glVertex3d (vertex.x, vertex.y, vertex.z);
	  
	  vertex = next_vertex (i);
	  glNormal3d (mp_iterator->normal.x, 
				  mp_iterator->normal.y, 
				  mp_iterator->normal.z);
	  
	  tex_x = (tex_width > 0.0) ?
		mp_iterator->texture_coordinates.x / tex_width : 1.0;
	  glTexCoord2d (tex_x, tex_y);
	  glVertex3d (vertex.x, vertex.y, vertex.z);
	} 

  glEnd ();

  m_texture_offsets [i] = tex_y;
}

// Return the corrdinates for the next vertex to be plotted.  Also,
// calculate the texture coordinates and store them in the iterator.
// Any neccary state information is also stored in the iterator.
Three_Vector 
Gl_Road_Segment::next_vertex (int strip)
{
  assert (mp_elevation_curve != 0);

  // Remember the last coordinates.
  Three_Vector last_coords = mp_iterator->coordinates;
  double last_tex_dist = mp_iterator->texture_coordinates.y;
  //  double left_kerb_width = mp_left_kerb->width ();
  //  double right_kerb_width = mp_right_kerb->width ();
  // True if we've reached the end of the strip.
  bool left_start_transition = false;
  bool left_end_transition = false;
  bool right_start_transition = false;
  bool right_end_transition = false;

  // mp_iterator->first is true if we're starting a strip.
  if (mp_iterator->first)
	{
	  glEnd ();
	  glBegin (GL_QUAD_STRIP);
	  //glBegin (GL_LINE_STRIP);

	  if (strip == LEFT_KERB)
		{
		  // Bail out if there's no kerb.
		  if (mp_left_kerb->transition_end () == 0.0)
			{
              mp_iterator->last = true;
			  return mp_iterator->coordinates;
			}
		  mp_iterator->distance = mp_left_kerb->transition_start ();
		}
	  else if (strip == RIGHT_KERB)
		{
		  // Bail out if there's no kerb.
		  if (mp_right_kerb->transition_end () == 0.0)
			{
			  mp_iterator->last = true;
			  return mp_iterator->coordinates;
			}
		  mp_iterator->distance = mp_right_kerb->transition_start ();
		}
	  else
		{
		  mp_iterator->distance = 0.0;
		}
	  last_tex_dist = 0.0;
	  mp_iterator->texture_coordinates.y = 0.0;
	}
  else if (mp_iterator->left_side)
	{
	  // Clamp to the end of the segment.
	  if ((m_length - mp_iterator->distance - mp_iterator->resolution) 
		  < (0.01 * mp_iterator->resolution))
		{
		  mp_iterator->distance = m_length;
		  mp_iterator->last = true;
		}
 	  else if (strip == LEFT_KERB)
 		{
 		  increment_kerb_distance (*mp_left_kerb);
 		}
 	  else if (strip == RIGHT_KERB)
 		{
 		  increment_kerb_distance (*mp_right_kerb);
 		}
	  else
		{
		  mp_iterator->distance += mp_iterator->resolution;
		}
	}

  size_t substrips = 1;
  if (strip == LEFT_KERB)
	{
	  mp_left_kerb->length (m_length);
	  substrips = mp_left_kerb->substrips ();
	  if (mp_iterator->distance < mp_left_kerb->start ())
		{
		  left_start_transition = true;
		}
	  else if (mp_iterator->distance > mp_left_kerb->end ())
		{
		  left_end_transition = true;
		}
	}
  else if (strip == RIGHT_KERB)
	{
	  mp_right_kerb->length (m_length);
	  substrips = mp_right_kerb->substrips ();
	  if (mp_iterator->distance < mp_right_kerb->start ())
		{
		  right_start_transition = true;
		}
	  else if (mp_iterator->distance > mp_right_kerb->end ())
		{
		  right_end_transition = true;
		}
	}

  // Calculate the geometry of the strip.
  double across = 0.0;
  double up = 0.0;
  switch (strip)
	{
	case LEFT_BARRIER:
	  across = left_width (mp_iterator->distance); 
	  if (mp_iterator->left_side)
		{
		  up = m_left_wall_height;
		}
	  else
		{
		  up = 0.0;
		}
	  mp_iterator->texture_coordinates.x = up;
	  mp_iterator->normal = barrier_normal (mp_iterator->distance, across);
	  break;
	case LEFT_SHOULDER:
	  up = 0.0;
	  if (mp_iterator->left_side)
		{
		  across = left_width (mp_iterator->distance);
		}
	  else
		{
		  across = left_road_width (mp_iterator->distance);
		}
	  mp_iterator->texture_coordinates.x = across;
	  mp_iterator->normal = normal (mp_iterator->distance, across);
	  break;
	case LEFT_KERB:
	  up = 0.0;
	  across = 0.0;
	  if (!mp_iterator->left_side)
		{
		  up = mp_left_kerb->point (mp_iterator->substrip).y;
		  across = mp_left_kerb->point (mp_iterator->substrip).x;
		}
	  else
		{
		  up = mp_left_kerb->point (mp_iterator->substrip + 1).y;
		  across = mp_left_kerb->point (mp_iterator->substrip + 1).x;
		}
	  if (left_start_transition)
		{
		  up = 0.0;
		  across *= (mp_left_kerb->start_transition_width ()
					 / mp_left_kerb->width ());
		}
	  else if (left_end_transition)
		{
		  up = 0.0;
		  across *= (mp_left_kerb->end_transition_width ()
					 / mp_left_kerb->width ());
		}

	  across += left_road_width (mp_iterator->distance);
	  mp_iterator->texture_coordinates.x = across;
	  break;
	case TRACK:
	  up = 0.0;
	  if (mp_iterator->left_side)
		{
		  across = left_road_width (mp_iterator->distance);
		}
	  else
		{
		  across = -right_road_width (mp_iterator->distance);
		}
	  mp_iterator->texture_coordinates.x 
		= across + left_road_width (mp_iterator->distance);
	  mp_iterator->normal = normal (mp_iterator->distance, across);
	  break;
	case RIGHT_KERB:
	  up = 0.0;
	  across = 0.0;
	  if (mp_iterator->left_side)
		{
		  up = mp_right_kerb->point (mp_iterator->substrip).y;
		  across = -mp_right_kerb->point (mp_iterator->substrip).x;
		}
	  else
		{
		  up = mp_right_kerb->point (mp_iterator->substrip + 1).y;
		  across = -mp_right_kerb->point (mp_iterator->substrip + 1).x;
		}
	  if (right_start_transition)
		{
		  up = 0.0;
		  across *= (mp_right_kerb->start_transition_width ()
					 / mp_right_kerb->width ());
		}
	  else if (right_end_transition)
		{
		  up = 0.0;
		  across *= (mp_right_kerb->end_transition_width ()
					 / mp_right_kerb->width ());
		}

	  across -= right_road_width (mp_iterator->distance);
	  mp_iterator->texture_coordinates.x = across;
	  break;
	case RIGHT_SHOULDER:
	  up = 0.0;
	  if (mp_iterator->left_side)
		{
		  across = -right_road_width (mp_iterator->distance);
		}
	  else
		{
		  across = -right_width (mp_iterator->distance);
		}
	  mp_iterator->texture_coordinates.x = across;
	  mp_iterator->normal = normal (mp_iterator->distance, across);
	  break;
	case RIGHT_BARRIER:
	  across = -right_width (mp_iterator->distance);
	  if (mp_iterator->left_side)
		{
		  up = 0.0;
		}
	  else
		{
		  up = m_right_wall_height;
		}
	  mp_iterator->texture_coordinates.x = up;
	  mp_iterator->normal = barrier_normal (mp_iterator->distance, across);
	  break;
	default:
	  assert (false);
	}

  if ((mp_iterator->last && m_last_segment)
	  && (strip != LEFT_KERB)
	  && (strip != RIGHT_KERB))
	{
	  // Return to the origin if we're at the end of the track.
	  mp_iterator->coordinates = 
        Three_Vector (-across * sin (mp_iterator->angle), 
                      across * cos (mp_iterator->angle), up);
	}
  else
	{
	  if (m_radius == 0.0)
		{
          const double extra = across 
            * (m_start_skew
               + ((m_end_skew - m_start_skew) * mp_iterator->distance 
                  / m_length)); 

          mp_iterator->coordinates = m_start_coords 
            + (Three_Vector ((mp_iterator->distance + extra), across, 0.0).
               rotate (0.0, 0.0, mp_iterator->angle));
		}
	  else
		{
          const double beta = arc () / 2.0;
          const double radius = 
            m_radius - across * (1.0 + m_start_skew / tan (beta));

          const Three_Vector 
            center (m_center 
                    + (across * m_start_skew / sin (beta) 
                       * Three_Vector (sin (m_start_angle + beta),
                                       -cos (m_start_angle + beta),
                                       0.0)));

		  mp_iterator->angle = 
			m_start_angle + mp_iterator->distance / m_radius;
          mp_iterator->coordinates = center 
            - (Three_Vector (0.0, radius, 0.0).
               rotate (0.0, 0.0, mp_iterator->angle));
		}

      mp_iterator->coordinates.z = mp_elevation_curve->
        interpolate (mp_iterator->distance + m_start_distance)
        + bank_height (mp_iterator->distance, across)
        + up;
	}

  if (!mp_iterator->first
	  && ((strip == LEFT_BARRIER) || (strip == RIGHT_BARRIER)))
	{
	  if (mp_iterator->left_side)
		{
		  double dx = mp_iterator->coordinates.x - last_coords.x;
		  double dy = mp_iterator->coordinates.y - last_coords.y;
		  double dist = std::sqrt (dx*dx + dy*dy);
		  mp_iterator->texture_coordinates.y = last_tex_dist + dist; 
		}
	}
  else
	{
	  mp_iterator->texture_coordinates.y = mp_iterator->distance;
	}

  if (!mp_iterator->left_side && mp_iterator->last)
	{
	  mp_iterator->distance = 0.0;
	}
  
  mp_iterator->first = false;

  if (mp_iterator->last && !mp_iterator->left_side)
	{
	  mp_iterator->substrip++;
	  mp_iterator->last = mp_iterator->substrip == substrips;
	  mp_iterator->first = !mp_iterator->last;
	}

  // Set mp_iterator->left_side to the correct value for the next time
  // next_vertex() is called.
  mp_iterator->left_side = !mp_iterator->left_side;

  // Update the track's bounds.
  m_bounds.enclose (Rectangle (mp_iterator->coordinates.x,
                               mp_iterator->coordinates.y,
                               mp_iterator->coordinates.x,
                               mp_iterator->coordinates.y));

  return mp_iterator->coordinates;
}

void
Gl_Road_Segment::draw () const
{
  glCallList (m_gl_list_id);
}

void
Gl_Road_Segment::set_braking_marker (Braking_Marker* marker)
{
  m_braking_markers.push_back (marker);
}

void 
Gl_Road_Segment::add_model_info (const Model_Info& info)
{
  m_models.push_back (info);
}

void
Gl_Road_Segment::set_start (const Three_Vector& start_coords, 
                            double start_distance,
                            double start_angle,
                            double start_bank,
                            const std::vector <double>& texture_offsets)
{
  Road_Segment::set_start (start_coords, 
                           start_distance,
                           start_angle, 
                           start_bank,
                           texture_offsets);
  m_texture_offsets = texture_offsets;
}

void 
Gl_Road_Segment::increment_kerb_distance (const Kerb& kerb)
{
  // Go to the end of the kerb if we were at the end transition.
  if ((mp_iterator->distance >= kerb.end ())
	  && ((kerb.end () - mp_iterator->distance) 
		  < (0.01 * mp_iterator->resolution)))
	{
	  mp_iterator->distance = kerb.transition_end ();
	  mp_iterator->last = true;
	}
  else if (mp_iterator->distance < kerb.start ())
	{
	  mp_iterator->distance = kerb.start ();
	}
  else if ((mp_iterator->distance + mp_iterator->resolution) 
		   >= kerb.transition_end ())
	{
	  mp_iterator->distance = kerb.end ();
	}
  else
	{
	  mp_iterator->distance += mp_iterator->resolution;
	}
}
