#include <cppunit/TestAssert.h>
#include <cppunit/TestCaller.h>
#include <cppunit/TestSuite.h>
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/extensions/TestFactoryRegistry.h>

#include <../include/vamos/track/Road_Segment.h>
#include <../include/vamos/geometry/Conversions.h>

using namespace Vamos_Geometry;
using namespace Vamos_Track;
using namespace std;

bool vector_equal (const Three_Vector& v1, const Three_Vector& v2)
{
  return abs (v1.x - v2.x) < 1e-6
    && abs (v1.y - v2.y) < 1e-6
    && abs (v1.z - v2.z) < 1e-6;
}

class Straight : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE (Straight);
  CPPUNIT_TEST (start);
  CPPUNIT_TEST (end);
  CPPUNIT_TEST (off_start);
  CPPUNIT_TEST (off_end);
  CPPUNIT_TEST_SUITE_END ();

  Road_Segment* segment;
  Spline* elevation;

public:
  void setUp ()
  {
    segment = new Road_Segment (100.0, 0.0, 10.0, 10.0, 5.0, 5.0);
    segment->set_elevation_point (0.0, 0.0);
    segment->set_elevation_point (100.0, 10.0);
    elevation = new Spline ();
    segment->build_elevation (elevation, 1.0);
  }

  void tearDown ()
  {
    delete segment;
    delete elevation;
  }

  void start ()
  {
    Three_Vector world_pos (0.0, 0.0, 0.0);
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-2); // big tolerance
  }

  void end ()
  {
    Three_Vector world_pos (segment->length (), 0.0, 0.0);
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, segment->length (), 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 10.0, 1e-2); // big tolerance
  }

  void off_start ()
  {
    Three_Vector world_pos (-1.0, 1.0, 0.0);
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, -1.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, -1.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 1.0, 1e-6);
    // Don't know what to expect for elevation here.
  }

  void off_end ()
  {
    Three_Vector world_pos (segment->length () + 1, 1.0, 0.0);
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 1.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, segment->length () + 1, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 1.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 10.0, 1e-6);
  }
};

class Moved_Straight : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE (Moved_Straight);
  CPPUNIT_TEST (start);
  CPPUNIT_TEST (end);
  CPPUNIT_TEST (position);
  CPPUNIT_TEST_SUITE_END ();

  Road_Segment* segment;
  Spline* elevation;

public:
  void setUp ()
  {
    segment = new Road_Segment (100.0, 0.0, 10.0, 10.0, 5.0, 5.0);
    elevation = new Spline ();
    segment->build_elevation (elevation, 0.0);
    segment->set_start (Three_Vector (1.0, 2.0, 3.0),
                        4.0, pi/2.0, -pi/4.0, std::vector <double> ());
  }

  void tearDown ()
  {
    delete segment;
    delete elevation;
  }

  void start ()
  {
    Three_Vector world_pos (1.0, 2.0, 3.0);
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void end ()
  {
    Three_Vector world_pos (1.0, segment->length () + 2.0, 3.0);
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, segment->length (), 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void position ()
  {
    Three_Vector world_pos = segment->position (100.0, 100.0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (world_pos.x, -99.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (world_pos.y, 102.0, 1e-6);
  }
};

class Left_Turn : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE (Left_Turn);
  CPPUNIT_TEST (no_change);
  CPPUNIT_TEST (change_length);
  CPPUNIT_TEST (change_radius);
  CPPUNIT_TEST (change_arc);
  CPPUNIT_TEST (scale);
  CPPUNIT_TEST (start);
  CPPUNIT_TEST (end);
  CPPUNIT_TEST (start_left);
  CPPUNIT_TEST (start_right);
  CPPUNIT_TEST (end_left);
  CPPUNIT_TEST (end_right);
  CPPUNIT_TEST (position);
  CPPUNIT_TEST_SUITE_END ();

  Road_Segment* segment;
  double initial_length;
  double initial_radius;
  double initial_arc;
  Spline* elevation;

public:
  void setUp ()
  {
    initial_length = pi * 50.0;
    initial_radius = 100.0;
    initial_arc = initial_length / initial_radius;

    segment = new Road_Segment (initial_length, initial_radius, 
                                10.0, 10.0, 5.0, 5.0);
    segment->set_width_point (initial_length, 30.0, LEFT);
    segment->set_width_point (initial_length / 2.0, 20.0, RIGHT);
    segment->set_road_width_point (initial_length, 1.0, LEFT);
    segment->set_road_width_point (initial_length / 2.0, 10.0, RIGHT);
    segment->set_elevation_point (initial_length / 2.0, 10.0);

    elevation = new Spline ();
    segment->build_elevation (elevation, 0.0);
    segment->set_start (Three_Vector (100.0, 2.0, 3.0),
                        111, pi / 2.0, 0.5, std::vector <double> ());
  }
  void tearDown ()
  {
    delete elevation;
    delete segment;
  }

  void no_change ()
  {
    CPPUNIT_ASSERT_EQUAL (segment->length (), initial_length);
    CPPUNIT_ASSERT_EQUAL (segment->radius (), initial_radius);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->length () / segment->radius (),
                                  initial_arc, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->left_width (initial_length / 4.0), 
                                  15.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->right_width (initial_length / 2.0), 
                                  20.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->
                                  left_road_width (initial_length / 4.0), 
                                  4.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->right_road_width (initial_length), 
                                  10.0, 1e-6);
    CPPUNIT_ASSERT_EQUAL (segment->start_coords(), 
                          Three_Vector (100.0, 2.0, 3.0));
    CPPUNIT_ASSERT (vector_equal (segment->end_coords(), 
                                  Three_Vector (0.0, 102.0, 0.0)));
  }

  void change_length ()
  {
    segment->set_length (2.0 * initial_length);

    CPPUNIT_ASSERT_EQUAL (segment->length (), 2.0 * initial_length);
    CPPUNIT_ASSERT_EQUAL (segment->radius (), initial_radius);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->length () / segment->radius (), 
                                  initial_arc * 2.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->left_width (initial_length / 2.0), 
                                  15.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->right_width (initial_length), 
                                  20.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->
                                  left_road_width (initial_length / 2.0), 
                                  4.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->
                                  right_road_width (initial_length * 2.0), 
                                  10.0, 1e-6);
  }

  void change_radius ()
  {
    segment->set_radius (2.0 * initial_radius);

    CPPUNIT_ASSERT_EQUAL (segment->length (), initial_length);
    CPPUNIT_ASSERT_EQUAL (segment->radius (), 2.0 * initial_radius);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->length () / segment->radius (), 
                                  initial_arc / 2.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->left_width (initial_length / 4.0), 
                                  15.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->right_width (initial_length / 2.0), 
                                  20.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->
                                  left_road_width (initial_length / 4.0), 
                                  4.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->right_road_width (initial_length), 
                                  10.0, 1e-6);
  }

  void change_arc ()
  {
    segment->set_arc (initial_arc / 2.0);

    CPPUNIT_ASSERT_EQUAL (segment->length (), initial_length / 2);
    CPPUNIT_ASSERT_EQUAL (segment->radius (), initial_radius);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->length () / segment->radius (), 
                                  initial_arc / 2.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->left_width (initial_length / 8.0), 
                                  15.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->right_width (initial_length / 2.0), 
                                  20.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->
                                  left_road_width (initial_length / 8.0), 
                                  4.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->
                                  right_road_width (initial_length / 2.0), 
                                  10.0, 1e-6);
  }

  void scale ()
  {
    segment->scale (0.5);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->length (), 
                                  initial_length / 2.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->radius (), 
                                  initial_radius / 2.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (segment->arc (), initial_arc, 1e-6);
  }

  void start ()
  {
    Three_Vector world_pos (segment->start_coords ());
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 10.0, 1e-6);
  }

  void start_left ()
  {
    Three_Vector world_pos (segment->start_coords ());
    world_pos.x -= 10.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 10.0, 1e-6);
  }

  void start_right ()
  {
    Three_Vector world_pos (segment->start_coords ());
    world_pos.x += 10.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, -10.0, 1e-6);
  }

  void end ()
  {
    Three_Vector world_pos (segment->end_coords ());
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 50 * pi, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 10.0, 1e-6);   
  }

  void end_left ()
  {
    Three_Vector world_pos (segment->end_coords ());
    world_pos.y -= 10.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 50.0 * pi, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 10.0, 1e-6);
  }

  void end_right ()
  {
    Three_Vector world_pos (segment->end_coords ());
    world_pos.y += 10.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 50.0 * pi, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, -10.0, 1e-6);
  }

  void position ()
  {
    Three_Vector world_pos = segment->position (pi * 25.0, 50.0);
    const double r = 50.0 / sqrt (2.0);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (world_pos.x, r, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (world_pos.y, r + 2.0, 1e-6);
  }
};

class Right_Turn : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE (Right_Turn);
  CPPUNIT_TEST (start);
  CPPUNIT_TEST (start_left);
  CPPUNIT_TEST (start_right);
  CPPUNIT_TEST (end);
  CPPUNIT_TEST_SUITE_END ();

  Road_Segment* segment;
  Spline* elevation;

public:
  void setUp ()
  {
    segment = new Road_Segment (pi * 50.0, -100.0,
                                10.0, 10.0, 5.0, 5.0);
    elevation = new Spline ();
    segment->build_elevation (elevation, 0.0);
    segment->set_start (Three_Vector (100.0, 2.0, 3.0),
                        111, pi / 2.0, 0.5, std::vector <double> ());
  }

  void tearDown ()
  {
    delete elevation;
    delete segment;
  }

  void start ()
  {
    Three_Vector world_pos (segment->start_coords ());
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void start_left ()
  {
    Three_Vector world_pos (segment->start_coords ());
    world_pos.x -= 10.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 10.0, 1e-6);
  }

  void start_right ()
  {
    Three_Vector world_pos (segment->start_coords ());
    world_pos.x += 10.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, -10.0, 1e-6);
  }

  void end ()
  {
    Three_Vector world_pos (segment->end_coords ());
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 50 * pi, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }
};

class Skewed : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE (Skewed);
  CPPUNIT_TEST (start_coordinates);
  CPPUNIT_TEST (start_left);
  CPPUNIT_TEST (start_right);
  CPPUNIT_TEST (end_left);
  CPPUNIT_TEST (end_right);
  CPPUNIT_TEST_SUITE_END ();

  Road_Segment* segment;
  Spline* elevation;

public:
  void setUp ()
  {
    segment = new Road_Segment (100.0, 0.0, 10.0, 10.0, 5.0, 5.0);
    segment->set_skew (-1.0);
    elevation = new Spline ();
    segment->build_elevation (elevation, 0.0);
  }

  void tearDown ()
  {
    delete elevation;
    delete segment;
  }

  void start_coordinates ()
  {
    Three_Vector world_pos;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void start_left ()
  {
    Three_Vector world_pos (-10.0, 10.0, 0.0);
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 10.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void start_right ()
  {
    Three_Vector world_pos (10.0, -10.0, 0.0);
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, -10.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void end_left ()
  {
    Three_Vector world_pos (110.0, 10.0, 0.0);
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 100.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 10.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void end_right ()
  {
    Three_Vector world_pos (90.0, -10.0, 0.0);
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 100.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, -10.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }
};

class Skewed_Left_Turn : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE (Skewed_Left_Turn);
  CPPUNIT_TEST (start);
  CPPUNIT_TEST (start_left);
  CPPUNIT_TEST (start_right);
  CPPUNIT_TEST (end);
  CPPUNIT_TEST (end_left);
  CPPUNIT_TEST (end_right);
  CPPUNIT_TEST_SUITE_END ();

  Road_Segment* segment;
  Spline* elevation;

public:
  void setUp ()
  {
    segment = new Road_Segment (pi * 50.0, 100.0,
                                10.0, 10.0, 5.0, 5.0);
    elevation = new Spline ();
    segment->build_elevation (elevation, 0.0);
    segment->set_start (Three_Vector (100.0, 2.0, 3.0),
                        111, pi / 2.0, 0.0, std::vector <double> ());
    segment->set_skew (-0.5);
  }

  void tearDown ()
  {
    delete elevation;
    delete segment;
  }

  void start ()
  {
    Three_Vector world_pos (segment->start_coords ());
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void start_left ()
  {
    Three_Vector world_pos (segment->start_coords ());
    world_pos.x -= 10.0;
    world_pos.y -= 5.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 10.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void start_right ()
  {
    Three_Vector world_pos (segment->start_coords ());
    world_pos.x += 10.0;
    world_pos.y += 5.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, -10.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void end ()
  {
    Three_Vector world_pos (segment->end_coords ());
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 50 * pi, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);   
  }

  void end_left ()
  {
    Three_Vector world_pos (segment->end_coords ());
    world_pos.x -= 5.0;
    world_pos.y -= 10.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 50.0 * pi, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 10.0, 1e-6);
  }

  void end_right ()
  {
    Three_Vector world_pos (segment->end_coords ());
    world_pos.x += 5.0;
    world_pos.y += 10.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 50.0 * pi, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, -10.0, 1e-6);
  }
};

class Skewed_Right_Turn : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE (Skewed_Right_Turn);
  CPPUNIT_TEST (start);
  CPPUNIT_TEST (start_left);
  CPPUNIT_TEST (start_right);
  CPPUNIT_TEST (end);
  CPPUNIT_TEST (end_left);
  CPPUNIT_TEST (end_right);
  CPPUNIT_TEST_SUITE_END ();

  Road_Segment* segment;
  Spline* elevation;

public:
  void setUp ()
  {
    segment = new Road_Segment (pi * 50.0, -100.0,
                                10.0, 10.0, 5.0, 5.0);
    elevation = new Spline ();
    segment->build_elevation (elevation, 0.0);
    segment->set_start (Three_Vector (100.0, 2.0, 3.0),
                        111, pi / 2.0, 0.0, std::vector <double> ());
    segment->set_skew (0.5);
  }

  void tearDown ()
  {
    delete elevation;
    delete segment;
  }

  void start ()
  {
    Three_Vector world_pos (segment->start_coords ());
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void start_left ()
  {
    Three_Vector world_pos (segment->start_coords ());
    world_pos.x -= 10.0;
    world_pos.y += 5.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 10.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void start_right ()
  {
    Three_Vector world_pos (segment->start_coords ());
    world_pos.x += 10.0;
    world_pos.y -= 5.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, -10.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);
  }

  void end ()
  {
    Three_Vector world_pos (segment->end_coords ());
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 50 * pi, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.z, 0.0, 1e-6);   
  }

  void end_left ()
  {
    Three_Vector world_pos (segment->end_coords ());
    world_pos.x -= 5.0;
    world_pos.y += 10.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 50.0 * pi, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, 10.0, 1e-6);
  }

  void end_right ()
  {
    Three_Vector world_pos (segment->end_coords ());
    world_pos.x += 5.0;
    world_pos.y -= 10.0;
    Three_Vector track_pos;
    double off = segment->coordinates (world_pos, track_pos);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (off, 0.0, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.x, 50.0 * pi, 1e-6);
    CPPUNIT_ASSERT_DOUBLES_EQUAL (track_pos.y, -10.0, 1e-6);
  }
};

CPPUNIT_TEST_SUITE_REGISTRATION (Straight);
CPPUNIT_TEST_SUITE_REGISTRATION (Moved_Straight);
CPPUNIT_TEST_SUITE_REGISTRATION (Left_Turn);
CPPUNIT_TEST_SUITE_REGISTRATION (Skewed);
CPPUNIT_TEST_SUITE_REGISTRATION (Right_Turn);
CPPUNIT_TEST_SUITE_REGISTRATION (Skewed_Left_Turn);
CPPUNIT_TEST_SUITE_REGISTRATION (Skewed_Right_Turn);

int main ()
{
  CppUnit::TextUi::TestRunner runner;
  runner.addTest (CppUnit::TestFactoryRegistry::getRegistry ().makeTest ());
  runner.run ();
  return 0;
}
